/*
    This file is part of BlueJoy.

    BlueJoy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BlueJoy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BlueJoy.  If not, see <http://www.gnu.org/licenses/>.
*/

package org.bluejoy.controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import org.bluejoy.model.BtDevice;
import org.bluejoy.model.CustomFileFilter;
import org.bluejoy.model.FileIO;
import org.bluejoy.model.Model;
import org.bluejoy.model.Model.DeviceStatus;
import org.bluejoy.model.Model.FileTransferMethod;
import org.bluejoy.view.FileTransferTxPanel;



public class FileTransferTxController implements ActionListener
{
	private FileTransferTxPanel fileTransferTxPanel;
	private Model model;
	
	//constructor
	public FileTransferTxController()
	{
		fileTransferTxPanel = null;
		model = null;
	}
	
	//init
	public void init(Model model, FileTransferTxPanel fileTransferTxPanel)
	{
		this.model = model;
		this.fileTransferTxPanel = fileTransferTxPanel;
	}
	
	
	//ActionListener interface
	public void actionPerformed(ActionEvent e)
	{
		BtDevice selectedDevice = model.getSelectedDevice();
		
		//Browse button
		if (e.getSource().equals(fileTransferTxPanel.getBrowseButton()))
		{
			JFileChooser fileChooser = new JFileChooser();
			fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileChooser.addChoosableFileFilter(new CustomFileFilter());
			
			int returnValue = fileChooser.showOpenDialog(fileTransferTxPanel);
			if (returnValue == JFileChooser.APPROVE_OPTION)
			{
				File file = fileChooser.getSelectedFile();
				selectedDevice.setTxLocalFile(file);
			}
			
			model.dataChanged();
		}
		//Scan button
		else if (e.getSource().equals(fileTransferTxPanel.getScanButton()))
		{
			//set busy status
			model.setDeviceStatus(DeviceStatus.FTP_TRANSFER);
			
			//clear tree data
			model.getSelectedDevice().clearRemoteFileSystemRoot();
			model.dataChanged();
			
			//scan
			model.browseRemoteDevice();
			
			//set idle status
			model.setDeviceStatus(DeviceStatus.IDLE);
			model.dataChanged();
		}
		//TX FTP radio button
		else if (e.getSource().equals(fileTransferTxPanel.getFtpTxRadioButton()))
		{
			selectedDevice.setFileTransferTxMethod(FileTransferMethod.FTP);
			model.dataChanged();
		}
		//TX OPP radio button
		else if (e.getSource().equals(fileTransferTxPanel.getOppTxRadioButton()))
		{
			selectedDevice.setFileTransferTxMethod(FileTransferMethod.OPP);
			model.dataChanged();
		}
		//TX SPP radio button
		else if (e.getSource().equals(fileTransferTxPanel.getSppTxRadioButton()))
		{
			selectedDevice.setFileTransferTxMethod(FileTransferMethod.SPP);
			model.dataChanged();
		}
		//Start Transfer button
		else if (e.getSource().equals(fileTransferTxPanel.getStartTransferButton()))
		{
			//validate the local file path
			String localPath = fileTransferTxPanel.getLocalTextField().getText();
			if (FileIO.validateFile(localPath))
			{
				selectedDevice.setTxLocalFile(new File(localPath));
			}
			else
			{
				JOptionPane.showMessageDialog(fileTransferTxPanel,
						"Warning: Local file path is invalid.",
						"Invalid File",
						JOptionPane.WARNING_MESSAGE,
						null);
				return;
			}
			
			//check that a remote destination has actually been chosen for FTP
			if (model.getSelectedDevice().getFileTransferTxMethod().equals(FileTransferMethod.FTP))
			{
				if (fileTransferTxPanel.getRemoteTree().isSelectionEmpty())
				{
					JOptionPane.showMessageDialog(fileTransferTxPanel,
							"Warning: You must select a directory on the remote BT device for transfer.",
							"Invalid Remote Directory Selection",
							JOptionPane.WARNING_MESSAGE,
							null);
					return;
				}
			}
			
			//set busy status
			model.setDeviceStatus(DeviceStatus.FTP_TRANSFER);	//sending data
			model.dataChanged();
			
			//send data
			if (selectedDevice.getFileTransferTxMethod().equals(FileTransferMethod.FTP))
			{
				Object remotePathArray[] = fileTransferTxPanel.getRemoteTree().getSelectionPath().getPath();
				String remotePath = "/";
				if (remotePathArray.length > 1)
				{
					for (int i = 1; i < remotePathArray.length; i++)
						remotePath += remotePathArray[i] + "/";
				}
				remotePath = remotePath.substring(0, remotePath.length() - 1);	//remove '/' at the end of the string 
				selectedDevice.setTxTargetPath(remotePath);
				model.sendFileUsingFTP();
			}
			else if (selectedDevice.getFileTransferTxMethod().equals(FileTransferMethod.OPP))
			{
				model.sendFileUsingOPP();
			}
			else if (selectedDevice.getFileTransferTxMethod().equals(FileTransferMethod.SPP))
			{
				//do nothing; not implemented yet...
			}
			
			//set status idle
			model.setDeviceStatus(DeviceStatus.IDLE);
			model.dataChanged();
		}
	}
}
