/*
    This file is part of BlueJoy.

    BlueJoy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BlueJoy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BlueJoy.  If not, see <http://www.gnu.org/licenses/>.
*/

package org.bluejoy.model;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import javax.bluetooth.DataElement;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;

import org.bluejoy.CustomLogger;
import org.bluejoy.CustomLogger.Level;
import org.bluejoy.model.Model.FileTransferMethod;



public class BtDevice
{
	//get a handle on logger
	private CustomLogger logger;
	
	//obtained from DiscoveryAgent
	private RemoteDevice remoteDevice;
	private DeviceClass cod;
	private ServiceRecord[] serviceRecord;
	
	//obtained from RemoteDevice
	private String deviceName;
	private String deviceAddress;
	
	//obtained from ServiceRecord
	private ArrayList<ServiceProfile> services;
	
	//device-specific state variables
	//receive
	private File rxLocalFile;
	private String rxTargetName;
	private String rxTargetPath;
	private FileTransferMethod fileTransferRxMethod;
	
	//send
	private File txLocalFile;
	private String txTargetName;
	private String txTargetPath;
	private FileTransferMethod fileTransferTxMethod;
	
	//common
	private RemoteTreeNode remoteFileSystemRoot;
	private ArrayList<FileTransferMethod> supportedMethods;
	
	
	//constructor
	public BtDevice(RemoteDevice remoteDevice, DeviceClass deviceClass)
	{
		logger = CustomLogger.getInstance();
		this.remoteDevice = remoteDevice;
		this.cod = deviceClass;
		serviceRecord = null;
		services = new ArrayList<ServiceProfile>();
		
		//state variables
		rxLocalFile = null;
		rxTargetName = null;
		rxTargetPath = null;
		fileTransferRxMethod = FileTransferMethod.FTP;
		
		txLocalFile = null;
		txTargetName = null;
		txTargetPath = null;
		fileTransferTxMethod = FileTransferMethod.OPP;
		
		remoteFileSystemRoot = null;
		supportedMethods = new ArrayList<FileTransferMethod>();
	}
	
	
	//reset root component to default state
	public void clearRemoteFileSystemRoot()
	{
		remoteFileSystemRoot = null;
	}
	
	
	//utilities
	public boolean isFileTransferSupported(FileTransferMethod method)
	{
		for (int i = 0; i < supportedMethods.size(); i++ )
		{
			if (supportedMethods.get(i).equals(method))
				return true;
		}
		
		return false;
	}
	
	
	public void parseRemoteDevice()
	{
		try
		{
			deviceName = remoteDevice.getFriendlyName(false);
			deviceAddress = remoteDevice.getBluetoothAddress();
		}
		catch (IOException ioe)
		{
			logger.log(Level.ERROR, "Exception thrown while parsing remote device.");
			ioe.printStackTrace();
		}
	}
	
	
	@SuppressWarnings("unchecked")
	public void parseServiceRecords()
	{
		if (serviceRecord == null)
			return;
		
		for (int i = 0; i < serviceRecord.length; i++)
		{
            String url = serviceRecord[i].getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);
            if (url == null)
            {
				continue;		//ignore blank URL and move on to the next one
            }
            
            int id[] = serviceRecord[i].getAttributeIDs();
			//for (int j = 0; j < id.length; j++)
            	//logger.log(Level.DEBUG, "Valid service record attribute ID = " + id[j]);
            
			DataElement serviceName = serviceRecord[i].getAttributeValue(0x0100);				//get name
			String name = null;
			if (serviceName != null)
				name = serviceName.getValue().toString();
			DataElement serviceClassIdList = serviceRecord[i].getAttributeValue(0x0001);		//get DATSEQ of UUIDs
			UUID serviceUUID = (UUID)( (Enumeration<DataElement>)serviceClassIdList.getValue() ).nextElement().getValue();	//get UUID - there should only be one
            
            if (serviceName != null)
            {
           		services.add(new ServiceProfile(name, serviceUUID, url));
                System.out.println("service " + serviceName.getValue() + " found " + url + "  using profile " + serviceUUID.toString());
            }
            else
            {
            	if (serviceUUID != null)
            		services.add(new ServiceProfile("", serviceUUID, url));
            	logger.log(Level.DEBUG, "service found " + url + "  using profile " + serviceUUID.toString());
            }
        }
	}
	
	
	//getters
	public RemoteDevice getRemoteDevice() {
		return remoteDevice;
	}
	
	public DeviceClass getCod() {
		return cod;
	}
	
	public String getDeviceName() {
		return deviceName;
	}

	public String getDeviceAddress() {
		return deviceAddress;
	}

	public ArrayList<ServiceProfile> getServices() {
		return services;
	}
	
	public File getTxLocalFile() {
		return txLocalFile;
	}
	
	public File getRxLocalFile() {
		return rxLocalFile;
	}
	
	public RemoteTreeNode getRemoteFileSystemRoot() {
		return remoteFileSystemRoot;
	}
	
	public FileTransferMethod getFileTransferTxMethod() {
		return fileTransferTxMethod;
	}

	public FileTransferMethod getFileTransferRxMethod() {
		return fileTransferRxMethod;
	}

	public String getRxTargetName() {
		return rxTargetName;
	}

	public String getRxTargetPath() {
		return rxTargetPath;
	}
	
	public String getTxTargetName() {
		return txTargetName;
	}

	public String getTxTargetPath() {
		return txTargetPath;
	}
	
	public ArrayList<FileTransferMethod> getSupportedMethods() {
		return supportedMethods;
	}
	
	
	//setters
	public void setServiceRecord(ServiceRecord[] serviceRecord) {
		this.serviceRecord = serviceRecord;
	}
	
	public void setTxLocalFile(File txLocalFile) {
		this.txLocalFile = txLocalFile;
	}
	
	public void setRxLocalFile(File rxLocalFile) {
		this.rxLocalFile = rxLocalFile;
	}
	
	public void setRemoteFileSystemRoot(RemoteTreeNode remoteFileSystemRoot) {
		this.remoteFileSystemRoot = remoteFileSystemRoot;
	}

	public void setFileTransferTxMethod(FileTransferMethod fileTransferTxMethod) {
		this.fileTransferTxMethod = fileTransferTxMethod;
	}

	public void setFileTransferRxMethod(FileTransferMethod fileTransferRxMethod) {
		this.fileTransferRxMethod = fileTransferRxMethod;
	}

	public void setRxTargetName(String rxTargetName) {
		this.rxTargetName = rxTargetName;
	}

	public void setRxTargetPath(String rxTargetPath) {
		this.rxTargetPath = rxTargetPath;
	}

	public void setTxTargetName(String txTargetName) {
		this.txTargetName = txTargetName;
	}

	public void setTxTargetPath(String txTargetPath) {
		this.txTargetPath = txTargetPath;
	}
}
