/*
    This file is part of BlueJoy.

    BlueJoy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BlueJoy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BlueJoy.  If not, see <http://www.gnu.org/licenses/>.
*/

package org.bluejoy.model;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.bluejoy.CustomLogger;
import org.bluejoy.CustomLogger.Level;


public class FileIO
{
	//returns the contents of the file in a byte array
    public static byte[] getBytesFromFile(File file)
    {
        InputStream is = null;
		try
		{
			is = new FileInputStream(file);
		}
		catch (FileNotFoundException fnfe)
		{
			CustomLogger.getInstance().log(Level.ERROR, "Exception thrown while opening FileInputStream");
			fnfe.printStackTrace();
			return null;
		}
		
		//get the size of the file
        long length = file.length();
    
        //You cannot create an array using a long type. It needs to be an int type.
        //Before converting to an int type, check to ensure that file is not larger than Integer.MAX_VALUE.
        if (length > Integer.MAX_VALUE)
        {
            //file is too large
        	CustomLogger.getInstance().log(CustomLogger.Level.ERROR, "The file " + file.getName() + " is too large.");
        	return null;
        }
    
        //create the byte array to hold the data
        byte[] bytes = new byte[(int)length];
    
        //read in the bytes
        int offset = 0;
        int numRead = 0;
        
        try
        {
			while (offset < bytes.length
			       && (numRead = is.read(bytes, offset, bytes.length-offset)) >= 0) {
			    offset += numRead;
			}
		}
        catch (IOException ioe)
        {
			ioe.printStackTrace();
			return null;
		}
	    
        //ensure all the bytes have been read in
        if (offset < bytes.length)
        {
        	CustomLogger.getInstance().log(CustomLogger.Level.ERROR, "Could not completely read file " + file.getName());
        	return null;
        }
	    
        //close the input stream and return bytes
        try
        {
			is.close();
		}
        catch (IOException ioe)
        {
        	CustomLogger.getInstance().log(Level.ERROR, "Exception thrown while closing input stream.");
			ioe.printStackTrace();
			return null;
		}
		
        return bytes;
    }
    
    
    //write byte array to file
    public static void writeToFile(ByteArrayOutputStream data, String filename)
    {
    	try
    	{
    		FileOutputStream out = new FileOutputStream(filename, false);		//over-write; no appending
    		data.writeTo(out);
    		out.flush();
    		out.close();
        }
    	catch (IOException ioe)
        {
        	CustomLogger.getInstance().log(CustomLogger.Level.ERROR, "Could not write byte data to file: " + filename);
        	ioe.printStackTrace();
        }
    }
    
    
	//verify that the given file exists on the local system
	public static boolean validateFile(String path)
	{
		File file = new File(path);
		boolean exists = file.exists();
		boolean isFile = file.isFile();
		
		if (exists && isFile)
        	return true;
		else
			return false;
    }
    
	//verify that the given directory exists on the local system
	public static boolean validateDirectory(String path)
	{
		File file = new File(path);
		boolean exists = file.exists();
		boolean isDirectory = file.isDirectory();
		
		if (exists && isDirectory)
        	return true;
		else
			return false;
    }
}
