/*
    This file is part of BlueJoy.

    BlueJoy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BlueJoy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BlueJoy.  If not, see <http://www.gnu.org/licenses/>.
*/

package org.bluejoy.model;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.List;
import java.util.Scanner;

import org.bluejoy.CustomLogger;
import org.bluejoy.CustomLogger.Level;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import javax.microedition.io.Connector;
import javax.obex.ClientSession;
import javax.obex.HeaderSet;
import javax.obex.Operation;
import javax.obex.ResponseCodes;


public class FtpClient
{
	private CustomLogger logger;
	private Model model;
	
	//File Browsing UUID = F9EC7BC4-953C-11D2-984E-525400DC9E09
	private byte[] fbUUID = {(byte)0xF9,(byte)0xEC,(byte)0x7B,(byte)0xC4,
			(byte)0x95,(byte)0x3C,(byte)0x11,(byte)0xD2,
			(byte)0x98,(byte)0x4E,(byte)0x52,(byte)0x54,
			(byte)0x00,(byte)0xDC,(byte)0x9E,(byte)0x09}; 
	
	//constructor
	public FtpClient()
	{
		//get a handle on logger
		logger = CustomLogger.getInstance();
		model = null;
	}
	
	public void init(Model model)
	{
		this.model = model;
	}

	
	//create a Client Session, connect to it, and enable file-browsing capabilities 
	private ClientSession initClientSession()
	{
		//verify that a BT device is selected
		if (model.getSelectedDevice() == null)
		{
			logger.log(Level.ERROR, "No Bluetooth device selected for FTP operation");
			return null;
		}
		
		//get the device URL of the File Transfer Profile for the selected device
		String deviceURL = null;
		for (int i = 0; i < model.getSelectedDevice().getServices().size(); i++)
		{
			if ( model.getSelectedDevice().getServices().get(i).getServiceUUID().equals(Model.OBEX_FILE_TRANSFER) )
			{
				deviceURL = model.getSelectedDevice().getServices().get(i).getServiceUrl();
				break;
			}
		}
		if (deviceURL == null)
		{
			logger.log(Level.ERROR, "FTP profile not found for selected device");
			return null;
		}

		//create a Client Session (i.e. socket) to the deviceURL
		ClientSession clientSession = null;
      	try
      	{
			clientSession = (ClientSession) Connector.open(deviceURL);
      	}
      	catch (IOException ioe)
      	{
      		logger.log(Level.ERROR, "Exception thrown while opening device URL = " + deviceURL);
			ioe.printStackTrace();
			return null;
		}
      	if (clientSession == null)
		{
      		logger.log(Level.ERROR, "Failed to open device URL = " + deviceURL);
			return null;
		}
      	
		//create header       
        HeaderSet header = clientSession.createHeaderSet();
        header.setHeader(HeaderSet.TARGET, fbUUID);		//set the TARGET to file browsing UUID
        try
        {
        	//connect to Client Session (i.e. socket)
			HeaderSet result = clientSession.connect(header);
			if (result.getResponseCode() != ResponseCodes.OBEX_HTTP_OK) {
				logger.log(Level.ERROR, "Failed to connect to client session");
				return null;
			}
		}
        catch (IOException ioe)
        {
        	logger.log(Level.ERROR, "Exception thrown while attempting to connect to client session");
			ioe.printStackTrace();
		}
		
      	return clientSession;
	}
	
	
	//set the path on remote BT device file system; path is set one directory at a time
	private int setRemotePath(ClientSession clientSession, String path, boolean backup)
	{
		//create header
		HeaderSet header = clientSession.createHeaderSet();
		header.setHeader(HeaderSet.NAME, path);	//"" = root
		try
		{
			//set path on remote BT device
			HeaderSet result = clientSession.setPath(header, backup, false);
			if (result.getResponseCode() != ResponseCodes.OBEX_HTTP_OK)
			{
				logger.log(Level.ERROR, "Failed to set path on remote BT device; path = " + path);
				return -1;
			}			
		}
		catch (IOException ioe)
		{
			logger.log(Level.ERROR, "Exception thrown while setting path on remote BT device; path = " + path);
			ioe.printStackTrace();
			return -1;
		}
		return 0;
	}
	
	
	//similar setRemotePath() mehtod; this method ONLY moves the directory path one level back (i.e. equivalent to "cd ..")
	private int setRemotePathBackward(ClientSession clientSession)
	{
		HeaderSet header = clientSession.createHeaderSet();
		try {
			HeaderSet result = clientSession.setPath(header, true, false);
			if (result.getResponseCode() != ResponseCodes.OBEX_HTTP_OK)
			{
				logger.log(Level.ERROR, "Failed to move path one level up on remote BT device");
				return -1;
			}			
		}
		catch (IOException ioe)
		{
			logger.log(Level.ERROR, "Exception thrown while moving path one level up on remote BT device");
			ioe.printStackTrace();
			return -1;
		}
		return 0;
	}
	
	
	//file browsing on remote BT device filesystem
	public void ftpBrowseAll()
	{
        //create and initialize Client session
        ClientSession clientSession = initClientSession();
        
      	//set the initial remote path; "" is the root on the remote BT device file system
      	if (setRemotePath(clientSession, "", false) == -1)
      	{
      		logger.log(Level.ERROR, "Failed to set path to root on remote BT device");
      		try
      		{
      			clientSession.disconnect(null);
      			clientSession.close();
      		}
      		catch (IOException ioe)
      		{
      			logger.log(Level.ERROR, "Exception thrown while setting remote path to root on remote BT device");
      			return;
      		}
      	}
      	
		//create root node for remote device file structure
		RemoteFile root = new RemoteFile("/", "", true);
		RemoteTreeNode rootNode = new RemoteTreeNode(root);
        
      	//recursively browse the remote directory structure, starting at the root node
      	this.recursiveBrowse(clientSession, rootNode);
      	
      	//save the root node and its children (i.e. basically the entire tree)
      	model.getSelectedDevice().setRemoteFileSystemRoot(rootNode);

      	//tear down client session
      	try
      	{
			clientSession.disconnect(null);
			clientSession.close();
		}
      	catch (IOException ioe)
      	{
			logger.log(Level.ERROR, "Exception thrown while tearing down client session");
			ioe.printStackTrace();
		}
	}
	
	
	private int recursiveBrowse(ClientSession clientSession, RemoteTreeNode parentNode)
	{
		//TODO - implement proper error-handling for this recursive method
		//currently, we just assume that errors never happen, and if they do, we press on
		
		//scan for all files and folders at the parent node; add all files/folders found to this node 
		browseRemoteDirectory(clientSession, parentNode);
		logger.log(Level.DEBUG, "parent child count = " + parentNode.getChildCount());
		
		//recursively browse each child directory found
		for (int i = 0; i < parentNode.getChildCount(); i++)
		{
			RemoteFile file = ((RemoteFile)((RemoteTreeNode)parentNode.getChildAt(i)).getUserObject());
			if (file.isDirectory())
			{
				//set path forward to the child directory
				setRemotePath(clientSession, file.getName(), false);
				
				//make recursive call
				recursiveBrowse(clientSession, (RemoteTreeNode)parentNode.getChildAt(i));
				
				//set path backward
				setRemotePathBackward(clientSession);
			}
		}
		
		return 0;
	}
	

	//get, parse, and store the file/folder listing from the remote BT device
	private void browseRemoteDirectory(ClientSession clientSession, RemoteTreeNode parent)
	{
		//TODO - reduce scope of error-handling try/catch blocks; add more try/catch blocks where needed
		//currently, too many things are done under one try/catch block
		
		//create header
		HeaderSet header = clientSession.createHeaderSet();
		header.setHeader(HeaderSet.TYPE, "x-obex/folder-listing");		//tell the remote BT device to provide folder/file listing for its PWD (present working directoy)
		logger.log(Level.DEBUG, "connection ID = " + clientSession.getConnectionID());

        //create PULL Operation
        Operation pullOperation = null;
        String xmlString = "";			//the complete XML string
		try {
			//retrieve the folder-listings in the form of an XML doc
			String xmlLine = "";		//a single XML line
			pullOperation = clientSession.get(header);
			BufferedReader in = new BufferedReader(new InputStreamReader(pullOperation.openInputStream()));
        	while( (xmlLine = in.readLine()) != null )
        	{
        		//filter out DOCTYPE and ATTLIST
        		Scanner scanXml = new Scanner(xmlLine.trim());
        		String scanString = scanXml.findInLine("(<!DOCTYPE)|(<!ATTLIST) .+");
        		if (scanString == null)
        		{
        			xmlString += xmlLine;
					logger.log(Level.DEBUG, xmlLine);
        		}
         	}
        	
        	//tear down the GET operation
        	in.close();
        	pullOperation.close();
        	
        	//use JDOM and SAX to parse the XML string; build XML doc
        	Document doc = null;
    		try
    		{
	    		SAXBuilder parser = new SAXBuilder();
	    		parser.setDTDHandler(null);
	    		doc = parser.build(new StringReader(xmlString));
    		}
    		catch(JDOMException jdome)
    		{
    			logger.log(Level.ERROR, "Exception thrown while building XML Document using JDOM and SAXBuilder");
    			jdome.printStackTrace();
    			return;
    		}
    		
    		//get the root
    		Element root = doc.getRootElement();
    		
    		//process the folder names and properties
    		List<?> folders = (List<?>)root.getChildren("folder");		//warning: unchecked cast
    		for (int i = 0; i < folders.size(); i++)
    		{
    			//read folder name
    			String name = ((Element)folders.get(i)).getAttributeValue("name");
    			
    			//set folder's path
    			String path;
    			if ( ((RemoteFile)parent.getUserObject()).getName().equals("/") )
    				path = "/";	//avoid "//" problem at the root
    			else			//otherwise, normal path construction
    				path = ((RemoteFile)parent.getUserObject()).getPath() + ((RemoteFile)parent.getUserObject()).getName() + "/";
    			
    			RemoteFile folder = new RemoteFile(name, path, true);
    			
    			//read folder created date
				String createdDate = ((Element)folders.get(i)).getAttributeValue("created");
				if (createdDate != null)
					folder.setDateCreated(createdDate);
				
				//read user permissions
				String permissions = ((Element)folders.get(i)).getAttributeValue("user-perm");
				if (permissions != null)
					folder.setPermissions(permissions);
				
				//read memory type
				String memoryType = ((Element)folders.get(i)).getAttributeValue("mem-type");
				if (memoryType != null)
					folder.setMemoryType(memoryType);
				
				//add folder to parent node
				//DefaultMutableTreeNode folderNode = new DefaultMutableTreeNode(folder);
				RemoteTreeNode folderNode = new RemoteTreeNode(folder);
				parent.add(folderNode);
    		}
    		
    		//process the file names and properties
    		List<?> files = (List<?>)root.getChildren("file");	//warning: unchecked cast
    		for (int i = 0; i < files.size(); i++)
    		{
    			//read file name
    			String name = ((Element)files.get(i)).getAttributeValue("name");
    			
    			//set file path
    			String path = ((RemoteFile)parent.getUserObject()).getPath() + ((RemoteFile)parent.getUserObject()).getName() + "/";
    			
    			RemoteFile file = new RemoteFile(name, path, false);
    			
    			//read file created date
    			String createdDate = ((Element)files.get(i)).getAttributeValue("modified");
				if (createdDate != null)
					file.setDateCreated(createdDate);
				
				//read user permissions
				String permissions = ((Element)files.get(i)).getAttributeValue("user-perm");
				if (permissions != null)
					file.setPermissions(permissions);
				
				//read memory type
				String memoryType = ((Element)files.get(i)).getAttributeValue("mem-type");
				if (memoryType != null)
					file.setMemoryType(memoryType);
				
				//read size
				String size = ((Element)files.get(i)).getAttributeValue("size");
				if (size != null)
					file.setSize(Long.parseLong(size));
				
				//add file to parent node
				RemoteTreeNode fileNode = new RemoteTreeNode(file);
				parent.add(fileNode);
    		}
        	
        }
		catch (Exception e)
        {
        	logger.log(Level.ERROR, "Exception thrown while creating GET operatiion, reading data stream, and parsing the XML results");
        	e.printStackTrace();
        }
	}

	
	//FTP file transfer methods
	public void ftpPush(File file, String targetPath)
	{
		//TODO - Proper error-handling
		//currently, no clean-up is performed if an error occurs
		
		//create and initialize Client session
        ClientSession clientSession = initClientSession();
        
		//load data to be sent
        byte data[] = FileIO.getBytesFromFile(file);

		//create header
        HeaderSet hsOperation = null;
		
		//prepare to set remote BT device path
		Scanner remotePathScanner = new Scanner(targetPath);
		remotePathScanner.useDelimiter("/");

		//set path to root
		int result = 0;
		result = setRemotePath(clientSession, "", false);
		if (result < 0)
			return;
		
		//set the path one folder at a time (i.e. repeatedly) until we reach the desired folder/depth 
		while (remotePathScanner.hasNext())
		{
			String remotePathValue = remotePathScanner.next();
			logger.log(Level.DEBUG, "remotePathValue = " + remotePathValue);
			
			result = setRemotePath(clientSession, remotePathValue, false);
			if (result < 0)
				return;
		}
		
		//set name and properties of the file to be pushed
		hsOperation = clientSession.createHeaderSet();
        hsOperation.setHeader(HeaderSet.NAME, file.getName());
        //hsOperation.setHeader(HeaderSet.TYPE, "x-obex/file");
        hsOperation.setHeader(HeaderSet.TYPE, "binary");
        hsOperation.setHeader(HeaderSet.LENGTH, (long)data.length);
        logger.log(Level.DEBUG, "NAME = " + file.getName() + "; LENGTH = " + (long)data.length);

        //create PUT Operation
        Operation putOperation;
		try
		{
			//send data
			putOperation = clientSession.put(hsOperation);
        	OutputStream os = putOperation.openOutputStream();
        	os.write(data);
        	os.close();
        	putOperation.close();
        }
		catch (Exception e)
		{
			logger.log(Level.ERROR, "Exception thrown while pushing data to remote BT device");
         	e.printStackTrace();
        }
        
		//tear down client session
		try
		{
			clientSession.disconnect(null);
            clientSession.close();
		}
		catch (IOException ioe)
		{
			logger.log(Level.ERROR, "Exception thrown while tearing down client session");
			ioe.printStackTrace();
		}
	}
	
	
	
	public void ftpPull(String targetName, String targetPath, String localPath)
	{
		//TODO - Proper error-handling
		//currently, no clean-up is performed if an error occurs
		
		//create and initialize Client session
        ClientSession clientSession = initClientSession();

        //create header
		HeaderSet hsOperation = null;
		
		//prepare to set path on remote BT device
		Scanner remotePathScanner = new Scanner(targetPath);
		remotePathScanner.useDelimiter("/");
		
		//set path to root
		int result = 0;
		result = setRemotePath(clientSession, "", false);
		if (result < 0)
			return;
		
		//set the path one folder at a time (i.e. repeatedly) until we reach the desired folder/depth 
		while (remotePathScanner.hasNext())
		{
			String remotePathValue = remotePathScanner.next();
			logger.log(Level.DEBUG, "remotePathValue = " + remotePathValue);
			
			result = setRemotePath(clientSession, remotePathValue, false);
			if (result < 0)
				return;
		}

      	//set name and properties of the file to be pulled
		hsOperation = clientSession.createHeaderSet();
        hsOperation.setHeader(HeaderSet.NAME, targetName);
        //hsOperation.setHeader(HeaderSet.TYPE, "x-obex/file");
        //hsOperation.setHeader(HeaderSet.TYPE, "binary");
        //hsOperation.setHeader(HeaderSet.TYPE, "image/jpeg");

        //create GET Operation
        Operation getOperation = null;
		try
		{
			//receive data
			getOperation = clientSession.get(hsOperation);
	        InputStream in = getOperation.openInputStream();
			ByteArrayOutputStream out = new ByteArrayOutputStream();
	        
			int data = in.read();
	        while (data != -1) {
				out.write(data);
				data = in.read();
	        }

	        in.close();
	        getOperation.close();
	        
	        FileIO.writeToFile(out, localPath + File.separator + targetName);
	        out.close();
        }
		catch (Exception e)
		{
			logger.log(Level.ERROR, "Exception thrown while pulling data from remote BT device");
        	e.printStackTrace();
        }
        
		//tear down client session
		try
		{
			clientSession.disconnect(null);
            clientSession.close();
		}
		catch (IOException ioe) 
		{
			logger.log(Level.ERROR, "Exception thrown while tearing down client session");
			ioe.printStackTrace();
		}
	}
}
