/*
    This file is part of BlueJoy.

    BlueJoy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BlueJoy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with BlueJoy.  If not, see <http://www.gnu.org/licenses/>.
*/

package org.bluejoy.model;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Observable;
import java.util.Vector;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.UUID;

import org.bluejoy.CustomLogger;
import org.bluejoy.CustomLogger.Level;


public class Model extends Observable
{
	//Bluetooth profile codes
	static final UUID OBEX = new UUID(0x0008);
	static final UUID RFCOMM = new UUID(0x0003);
	static final UUID SERIAL_PORT = new UUID(0x1101);
	static final UUID OBEX_OBJECT_PUSH = new UUID(0x1105);
	static final UUID OBEX_FILE_TRANSFER = new UUID(0x1106);
	
	//major device classes
	static final int MAJOR_DEVICE_CLASS_MISCELLANEOUS = 0x0000;
	static final int MAJOR_DEVICE_CLASS_COMPUTER = 0x0100;
	static final int MAJOR_DEVICE_CLASS_PHONE = 0x0200;
	static final int MAJOR_DEVICE_CLASS_LAN = 0x0300;
	static final int MAJOR_DEVICE_CLASS_AUDIO_VIDIO = 0x0400;
	static final int MAJOR_DEVICE_CLASS_PERIPHERAL = 0x0500;
	static final int MAJOR_DEVICE_CLASS_IMAGING = 0x0600;
	static final int MAJOR_DEVICE_CLASS_UNCATEGORIZED = 0x1F00;
	
	//minor device classes for COMPUTER
	static final int MINOR_DEVICE_CLASS_DESKTOP = 0x04;
	static final int MINOR_DEVICE_CLASS_SERVER = 0x08;
	static final int MINOR_DEVICE_CLASS_LAPTOP = 0x0c;
	static final int MINOR_DEVICE_CLASS_HANDHELD_PDA = 0x10;
	static final int MINOR_DEVICE_CLASS_PALM_PDA = 0x14;
	static final int MINOR_DEVICE_CLASS_WEARABLE = 0x18;

	//minor device class for PHONE
	static final int MINOR_DEVICE_CLASS_CELLULAR = 0x04;
	static final int MINOR_DEVICE_CLASS_CORDLESS = 0x08;
	static final int MINOR_DEVICE_CLASS_SMARTPHONE = 0x0c;
	static final int MINOR_DEVICE_CLASS_MODEM = 0x10;
	static final int MINOR_DEVICE_CLASS_ISDN = 0x14;
	
	public enum DeviceStatus
	{
		UNINITIALIZED,
		IDLE,
		DEVICE_SEARCH,
		SERVICE_SEARCH,
		OPP_TRANSFER,
		FTP_TRANSFER,
		SPP_TRANSFER,
		DEVICE_NOT_PRESENT
	}
	
	public enum FileTransferMethod
	{
		OPP,
		FTP,
		SPP,
		NONE
	}

	private CustomLogger logger;
	private DeviceDiscovery deviceDiscovery;
	private ServiceDiscovery serviceDiscovery;
	private OppClient oppClient;
	private FtpClient ftpClient;
	
	//global state variables
	private HashMap<Integer, HashMap<Integer, String>> majorDeviceClass;
	private HashMap<Integer, String> minorDeviceClassComputer;
	private HashMap<Integer, String> minorDeviceClassPhone;
	private ArrayList<BtDevice> remoteDevices;
	private Vector<Vector<Object>> remoteDevicesTableData;
	private BtDevice selectedDevice;
	private int selectedDeviceTableRow;
	private DeviceStatus deviceStatus;	//i.e. local device status

	
	//constructor
	public Model()
	{
		//get a handle on logger
		logger = CustomLogger.getInstance();
        
		//initialize vars
		deviceDiscovery = new DeviceDiscovery();
		serviceDiscovery = new ServiceDiscovery();
		oppClient = new OppClient();
		ftpClient = new FtpClient();
        
		majorDeviceClass = new HashMap<Integer, HashMap<Integer, String>>();
		minorDeviceClassComputer = new HashMap<Integer, String>();
		minorDeviceClassPhone = new HashMap<Integer, String>();
		remoteDevices = new ArrayList<BtDevice>();
		remoteDevicesTableData = new Vector<Vector<Object>>();
		deviceStatus = DeviceStatus.UNINITIALIZED;
		selectedDevice = null;
		selectedDeviceTableRow = -1;
	}
	
	
	public void init()
	{
		deviceDiscovery.init(this);
		serviceDiscovery.init(this);
		oppClient.init(this);
		ftpClient.init(this);
		deviceStatus = DeviceStatus.IDLE;

		initHashMaps();
	}
	
	
	//init the device class hashmaps
	public void initHashMaps()
	{
		//add device classes to the hashmaps
		//add major devices first
		majorDeviceClass.put(MAJOR_DEVICE_CLASS_COMPUTER, minorDeviceClassComputer);
		majorDeviceClass.put(MAJOR_DEVICE_CLASS_PHONE, minorDeviceClassPhone);
		
		//minor device class for COMPUTER
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_DESKTOP, "Computer->Desktop");
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_SERVER, "Computer->Server");
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_LAPTOP, "Computer->Laptop");
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_HANDHELD_PDA, "Computer->Handhelp PDA");
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_PALM_PDA, "Computer->Palm PDA");
		minorDeviceClassComputer.put(MINOR_DEVICE_CLASS_WEARABLE, "Computer->Wearable");
		
		//minor device class for PHONE
		minorDeviceClassPhone.put(MINOR_DEVICE_CLASS_CELLULAR, "Phone->Cellular");
		minorDeviceClassPhone.put(MINOR_DEVICE_CLASS_CORDLESS, "Phone->Cordless");
		minorDeviceClassPhone.put(MINOR_DEVICE_CLASS_SMARTPHONE, "Phone->Smartphone");
		minorDeviceClassPhone.put(MINOR_DEVICE_CLASS_MODEM, "Phone->Modem");
		minorDeviceClassPhone.put(MINOR_DEVICE_CLASS_ISDN, "Phone->ISDN");
	}

	
	//indicate to observer that data has changed
	public void dataChanged()
	{
		setChanged();
		notifyObservers();
	}
	
	
	//reset all model components to default state
	public void clear()
	{
		remoteDevices.clear();
		deviceStatus = DeviceStatus.IDLE;
		remoteDevicesTableData.clear();
		selectedDevice = null;
		selectedDeviceTableRow = -1;
	}

	
    //scan methods
    public void scanForDevices()
    {
    	//scan for devices
		deviceDiscovery.startDeviceSearch();
		
		//parse the device info
		while (isDeviceSearchStarted())
		{
			try
			{
				Thread.sleep(1000);
			}
			catch (InterruptedException ie)
			{
				logger.log(Level.ERROR, "Exception thrown while waiting for device search to complete.");
				ie.printStackTrace();
			}
		}
		
		//for each device found, scan for services
		for (int i = 0; i < remoteDevices.size(); i++)
		{
			serviceDiscovery.startServiceSearch(remoteDevices.get(i).getRemoteDevice());
			while (isServiceSearchStarted())
			{
				try
				{
					Thread.sleep(1000);
				}
				catch (InterruptedException ie)
				{
					logger.log(Level.ERROR, "Exception thrown while waiting for service search to complete.");
					ie.printStackTrace();
				}
			}
		}
		
		//parse all the service records for each device
		for (int i = 0; i < remoteDevices.size(); i++)
			remoteDevices.get(i).parseServiceRecords();
	}
    
    
    //OPP data transfer methods
    public int sendFileUsingOPP()
	{
    	File txLocalFile = null;
    	if (selectedDevice != null)
    		txLocalFile = selectedDevice.getTxLocalFile();
    	
    	if (txLocalFile == null)
    		return -1;
    	oppClient.oppPush(txLocalFile);
    	return 0;
 	}
 
    
	//FTP data transfer methods
    public int sendFileUsingFTP()
	{
    	File txLocalFile = null;
    	String txTargetPath = null;
    	if (selectedDevice != null)
    	{
    		txLocalFile = selectedDevice.getTxLocalFile();
        	txTargetPath = selectedDevice.getTxTargetPath();
    	}
    	
    	if (txLocalFile == null || txTargetPath == null)
    		return -1;
    	ftpClient.ftpPush(txLocalFile, txTargetPath);
    	return 0;
	}
    
    
    public int receiveFileUsingFTP()
	{
    	File rxLocalFile = null;
    	String rxTargetName = null;
    	String rxTargetPath = null;
    	if (selectedDevice != null)
    	{
    		rxLocalFile = selectedDevice.getRxLocalFile();
    		rxTargetName = selectedDevice.getRxTargetName();
    		rxTargetPath = selectedDevice.getRxTargetPath();
    	}
    	
    	if (rxLocalFile == null)
    		return -1;
    	ftpClient.ftpPull(rxTargetName, rxTargetPath, rxLocalFile.getAbsolutePath());
    	return 0;
	}
    
    
    //scan directory structure on remote device
	public int browseRemoteDevice()
	{
		if (selectedDevice == null)
			return -1;
		ftpClient.ftpBrowseAll();
    	return 0;
	}
	
	
	//test if BT device is present on host
	public void checkBtDeviceOnHost()
	{
		try
		{
			LocalDevice.getLocalDevice();
		}
		catch (BluetoothStateException ste)
		{
			logger.log(Level.WARN, "No Bluetooth device detected on local system.");
			deviceStatus = DeviceStatus.DEVICE_NOT_PRESENT;
		}
		
		dataChanged();
	}
	
	
	//parse remote devices data; form a vector of vectors containing useful info to display in a table
	//also, determine which file transfer methods are supported by the remote BT device
	public void parseRemoteDevicesData()
	{
		Vector<Vector<Object>> dataVector = new Vector<Vector<Object>>();
		Vector<Object> rowVector = null;
		for (int i = 0; i < remoteDevices.size(); i++)
		{
			rowVector = new Vector<Object>();
			rowVector.add(new String(remoteDevices.get(i).getDeviceName()));
			rowVector.add(new String(remoteDevices.get(i).getDeviceAddress()));
			//rowVector.add(new String("Major = " + remoteDevices.get(i).getCod().getMajorDeviceClass() + 
			//		", Minor = " + remoteDevices.get(i).getCod().getMinorDeviceClass()));
			int major = remoteDevices.get(i).getCod().getMajorDeviceClass();
			int minor = remoteDevices.get(i).getCod().getMinorDeviceClass();
			rowVector.add(new String(majorDeviceClass.get(major).get(minor)));
			ArrayList<ServiceProfile> services = remoteDevices.get(i).getServices();
			String serviceString = new String();
			for (int j = 0; j < services.size(); j++)
			{
				//form string of names of all services offered 
				if (services.get(j).getServiceName().length() > 0)
					serviceString += services.get(j).getServiceName() + ", ";
				
				//form list of all file transfer services supported by BlueJoy
				if (services.get(j).getServiceUUID().equals(OBEX_OBJECT_PUSH))
					remoteDevices.get(i).getSupportedMethods().add(FileTransferMethod.OPP);
				else if (services.get(j).getServiceUUID().equals(OBEX_FILE_TRANSFER))
					remoteDevices.get(i).getSupportedMethods().add(FileTransferMethod.FTP);
				else if (services.get(j).getServiceUUID().equals(SERIAL_PORT))
					remoteDevices.get(i).getSupportedMethods().add(FileTransferMethod.SPP);
			}
			if (serviceString.lastIndexOf(',') != -1)
				rowVector.add(serviceString.substring(0, serviceString.lastIndexOf(',')));

			dataVector.add(rowVector);
		}
		
		remoteDevicesTableData = dataVector;
	}
	
	
	//select the default methods for RX and TX
	public void setPreferredMethods()
	{
		for (int i = 0 ; i < remoteDevices.size(); i++)
		{
			BtDevice device = remoteDevices.get(i);

			//RX
			//XXX - SPP is not yet implemented. So, choose the first method we find that isn't SPP
			FileTransferMethod rxMethod = FileTransferMethod.NONE;
			for (int j = 0; j < device.getSupportedMethods().size(); j++)
			{
				FileTransferMethod method = device.getSupportedMethods().get(j);
				if (method.equals(FileTransferMethod.OPP) == false && method.equals(FileTransferMethod.SPP) == false)
				{
					rxMethod = method;
					break;
				}
			}
			remoteDevices.get(i).setFileTransferRxMethod(rxMethod);
			
			//TX
			//XXX - SPP is not yet implemented. So, choose the first method we find that isn't SPP
			FileTransferMethod txMethod = FileTransferMethod.NONE;
			for (int j = 0; j < device.getSupportedMethods().size(); j++)
			{
				FileTransferMethod method = device.getSupportedMethods().get(j);
				if (method.equals(FileTransferMethod.SPP) == false)
				{
					txMethod = method;
					break;
				}
			}
			remoteDevices.get(i).setFileTransferTxMethod(txMethod);
		}
	}
	
	
	//getters
	public Vector<Vector<Object>> getRemoteDevicesTableData() {
		return remoteDevicesTableData;
	}
	
	public int getSelectedDeviceTableRow() {
		return selectedDeviceTableRow;
	}
	
	public ArrayList<BtDevice> getRemoteDevices() {
		return remoteDevices;
	}
	
	public boolean isDeviceSearchStarted() {
		return (deviceStatus == DeviceStatus.DEVICE_SEARCH);
	}

	public boolean isServiceSearchStarted() {
		return (deviceStatus == DeviceStatus.SERVICE_SEARCH);
	}
	
	public DeviceStatus getDeviceStatus() {
		return deviceStatus;
	}

	public BtDevice getSelectedDevice() {
		return selectedDevice;
	}
	
	
	//setters
	public void setSelectedDeviceTableRow(int selectedDeviceTableRow) {
		this.selectedDeviceTableRow = selectedDeviceTableRow;
	}
	
	public void setDeviceStatus(DeviceStatus deviceStatus) {
		this.deviceStatus = deviceStatus;
	}

	public void setSelectedDevice(String deviceName)
	{
		if (deviceName == null)
			selectedDevice = null;
		else
		{
			for (int i = 0; i < remoteDevices.size(); i++)
			{
				if (remoteDevices.get(i).getDeviceName().equals(deviceName))
					selectedDevice = remoteDevices.get(i);
			}
		}
	}
}
